import numpy as np

from ...dataclasses import TrainInfo


class TrainingLogger:
    def __init__(self, task: str, epochs: int) -> None:
        self.task: str = task
        self.epochs: int = epochs

        self.train_losses: list[float] = []
        self.train_accuracies: list[float] = []

        self.val_losses: list[float] = []
        self.val_accuracies: list[float] = []

        self.lr_values: list[float] = []

        self.avg_time_per_epoch: float = 0.0

    def log(self,
            train_loss: float,
            train_accuracy: float,
            val_loss: float,
            val_accuracy: float,
            lr: float,
            epoch: int,
            time_per_epoch: float,
            ) -> None:
        self.train_losses.append(train_loss)
        self.train_accuracies.append(train_accuracy) if self.task == "classification" else None

        self.val_losses.append(val_loss)
        self.val_accuracies.append(val_accuracy) if self.task == "classification" else None

        self.lr_values.append(lr)

        self.avg_time_per_epoch = (self.avg_time_per_epoch * (epoch - 1) + time_per_epoch) / epoch

    def get_info(self) -> TrainInfo:
        return TrainInfo(
            train_losses=np.array(self.train_losses),
            train_accuracies=np.array(self.train_accuracies) if self.task == "classification" else np.array([]),
            best_train_loss=np.min(self.train_losses),
            best_train_accuracy=np.max(self.train_accuracies) if self.task == "classification" else -np.inf,
            val_losses=np.array(self.val_losses),
            val_accuracies=np.array(self.val_accuracies) if self.task == "classification" else np.array([]),
            best_val_loss=np.min(self.val_losses),
            best_val_accuracy=np.max(self.val_accuracies) if self.task == "classification" else -np.inf,
            lr_values=np.array(self.lr_values),
        )

    def __repr__(self) -> str:
        epoch: int = len(self.train_losses)
        percentage: float = epoch / self.epochs
        percentage_str: str = f"{int(percentage * 100)}%" if percentage > 0.09 else f"0{int(percentage * 100)}%"
        epoch_msg: str = f"    -> epoch: {epoch}/{self.epochs} ({percentage_str})\n"

        # LOSS
        train_loss: str = f"{self.train_losses[-1]:.4f}"
        val_loss: str = f"{self.val_losses[-1]:.4f}"
        best_val_loss: str = f"{min(self.val_losses):.4f}"
        loss_msg: str = f"    -> loss: train({train_loss}), val({val_loss}), best({best_val_loss})\n"

        # ACCURACY
        train_accuracy: str = f"{self.train_accuracies[-1]:.4f}" if len(self.train_accuracies) > 0 else ""
        val_accuracy: str = f"{self.val_accuracies[-1]:.4f}" if len(self.val_accuracies) > 0 else ""
        best_val_accuracy: str = f"{max(self.val_accuracies):.4f}" if len(self.val_accuracies) > 0 else ""
        accuracy_msg: str = f"    -> accuracy: train({train_accuracy}), val({val_accuracy}), best({best_val_accuracy})\n" if len(self.train_accuracies) > 0 else ""

        # TIME
        time_remaining: int = int(self.avg_time_per_epoch * (self.epochs - len(self.train_losses)))
        time_remaining_str: str = f"{int(time_remaining/60)}m {time_remaining%60}s"
        time_msg: str = f"    -> time remaining: {time_remaining_str}"

        clear_line: str = "\x1b[2K" + "\r" + "\x1b[1A"

        msg: str = ""
        if epoch != 1:
            msg += clear_line*4 if self.task == "classification" else clear_line*3

        msg += epoch_msg + loss_msg + accuracy_msg + time_msg

        return msg



import numpy as np

from .transformation import Transformation


class Decimation(Transformation):
    def __init__(self, decimation_factor: int) -> None:
        """Initialize the Decimation object.

        Arguments
        ---------
        decimation_factor : int
            The factor by which to decimate the data.
        """
        self.decimation_factor: int = decimation_factor

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """Decimate the data.

        Arguments
        ---------
        x : np.ndarray
            The data to decimate.

        Returns
        -------
        np.ndarray
            The decimated data.
        """
        return x[::self.decimation_factor]



import pandas as pd

from catasta.models import TransformerRegressor, ApproximateGPRegressor
from catasta.archways import RegressionArchway
from catasta.transformations import (
    WindowSliding,
    Normalization,
    Decimation,
    Custom,
    Slicing,
)
from catasta.dataclasses import RegressionEvalInfo


def vanilla() -> None:
    path: str = "tests/models/"

    n_dim: int = 768
    model = TransformerRegressor(
        context_length=n_dim,
        n_patches=8,
        d_model=8,
        n_heads=4,
        n_layers=2,
        feedforward_dim=4,
        head_dim=4,
        dropout=0.0,
    )

    archway = RegressionArchway(
        model=model,
        path=path,
    )

    df = pd.read_csv("tests/data/nylon_elastic/strain/tri_30.csv")
    input = df["input"].to_numpy().flatten()
    output = df["output"].to_numpy().flatten()

    input_transformations = [
        Custom(lambda x: x[1_500_000: 2_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_transformations = [
        Custom(lambda x: x[1_500_000: 2_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    for t in input_transformations:
        input = t(input)
    for t in output_transformations:
        output = t(output)

    prediction = archway.predict(input)

    predicted_output = prediction.value
    true_output = output[-len(predicted_output):]
    true_input = input[-len(predicted_output):]

    info = RegressionEvalInfo(
        true_input=true_input,
        predicted_output=predicted_output,
        true_output=true_output,
    )

    print(info)


def vanilla_onnx() -> None:
    path: str = "tests/models/"

    n_dim: int = 768
    archway = RegressionArchway(
        path=path,
        from_onnx=True,
    )

    df = pd.read_csv("tests/data/nylon_elastic/strain/tri_30.csv")
    input = df["input"].to_numpy().flatten()
    output = df["output"].to_numpy().flatten()

    input_transformations = [
        Custom(lambda x: x[1_500_000: 2_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_transformations = [
        Custom(lambda x: x[1_500_000: 2_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    for t in input_transformations:
        input = t(input)
    for t in output_transformations:
        output = t(output)

    prediction = archway.predict(input)

    predicted_output = prediction.value
    true_output = output[-len(predicted_output):]
    true_input = input[-len(predicted_output):]

    info = RegressionEvalInfo(
        true_input=true_input,
        predicted_output=predicted_output,
        true_output=true_output,
    )

    print(info)


def gp() -> None:
    path: str = "tests/models/gp/"

    n_dim: int = 768
    model = ApproximateGPRegressor(
        context_length=n_dim,
        n_inducing_points=128,
    )

    archway = RegressionArchway(
        model=model,
        path=path,
    )

    df = pd.read_csv("tests/data/nylon_elastic/strain/tri_30.csv")
    input = df["input"].to_numpy().flatten()
    output = df["output"].to_numpy().flatten()

    input_transformations = [
        Custom(lambda x: x[1_500_000: 2_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_transformations = [
        Custom(lambda x: x[1_500_000: 2_000_000]),
        Normalization("minmax"),
        Decimation(decimation_factor=100),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    for t in input_transformations:
        input = t(input)
    for t in output_transformations:
        output = t(output)

    prediction = archway.predict(input)

    predicted_output = prediction.value
    true_output = output[-len(predicted_output):]
    true_input = input[-len(predicted_output):]

    info = RegressionEvalInfo(
        true_input=true_input,
        predicted_output=predicted_output,
        true_output=true_output,
    )

    print(info)


if __name__ == "__main__":
    # print("vanilla")
    # vanilla()
    # print("vanilla_onnx")
    # vanilla_onnx()
    # print("gp")
    gp()



import numpy as np
from scipy.signal import firwin, lfilter

from .transformation import Transformation


class FIRFiltering(Transformation):
    def __init__(self, fs: int, cutoff: float, numtaps: int) -> None:
        self.fs: int = fs
        self.cutoff: float = cutoff
        self.numtaps: int = numtaps

    def __call__(self, x: np.ndarray) -> np.ndarray:
        fir_coeff: np.ndarray = firwin(
            numtaps=self.numtaps,
            cutoff=self.cutoff,
            fs=self.fs,
            window="hamming",
        )
        return np.array(lfilter(fir_coeff, 1.0, x))



import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat, einsum

from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Sequential,
    ModuleList,
    Identity,
)
from torch.fft import fft2

from einops.layers.torch import Rearrange


class ResidualBlock(Module):
    def __init__(self, *,
                 d_model: int,
                 d_inner: int,
                 conv_bias: bool,
                 d_conv: int,
                 bias: bool,
                 d_state: int,
                 dt_rank: int,
                 ) -> None:
        super().__init__()
        self.mixer = MambaBlock(
            d_model=d_model,
            d_inner=d_inner,
            conv_bias=conv_bias,
            d_conv=d_conv,
            bias=bias,
            d_state=d_state,
            dt_rank=dt_rank,
        )
        self.norm = RMSNorm(d_model)

    def forward(self, x):
        output = self.mixer(self.norm(x)) + x

        return output


class MambaBlock(Module):
    def __init__(self, *,
                 d_model: int,
                 d_inner: int,
                 conv_bias: bool,
                 d_conv: int,
                 bias: bool,
                 d_state: int,
                 dt_rank: int,
                 ) -> None:
        super().__init__()

        self.d_inner = d_inner
        self.dt_rank = dt_rank

        self.in_proj = nn.Linear(d_model, d_inner * 2, bias=bias)

        self.conv1d = nn.Conv1d(
            in_channels=d_inner,
            out_channels=d_inner,
            bias=conv_bias,
            kernel_size=d_conv,
            groups=d_inner,
            padding=d_conv - 1,
        )

        self.x_proj = nn.Linear(d_inner, dt_rank + d_state * 2, bias=False)

        self.dt_proj = nn.Linear(dt_rank, d_inner, bias=True)

        A = repeat(torch.arange(1, d_state + 1), 'n -> d n', d=d_inner)
        self.A_log = nn.Parameter(torch.log(A))
        self.D = nn.Parameter(torch.ones(d_inner))
        self.out_proj = nn.Linear(d_inner, d_model, bias=bias)

    def forward(self, x: Tensor) -> Tensor:
        (b, l, d) = x.shape

        x_and_res = self.in_proj(x)  # shape (b, l, 2 * d_in)
        (x, res) = x_and_res.split(split_size=[self.d_inner, self.d_inner], dim=-1)

        x = rearrange(x, 'b l d_in -> b d_in l')
        x = self.conv1d(x)[:, :, :l]
        x = rearrange(x, 'b d_in l -> b l d_in')

        x = F.silu(x)

        y = self.ssm(x)

        y = y * F.silu(res)

        output = self.out_proj(y)

        return output

    def ssm(self, x):
        (d_in, n) = self.A_log.shape

        A = -torch.exp(self.A_log.float())  # shape (d_in, n)
        D = self.D.float()

        x_dbl = self.x_proj(x)  # (b, l, dt_rank + 2*n)

        (delta, B, C) = x_dbl.split(split_size=[self.dt_rank, n, n], dim=-1)  # delta: (b, l, dt_rank). B, C: (b, l, n)
        delta = F.softplus(self.dt_proj(delta))  # (b, l, d_in)

        y = self.selective_scan(x, delta, A, B, C, D)  # This is similar to run_SSM(A, B, C, u) in The Annotated S4 [2]

        return y

    def selective_scan(self, u, delta, A, B, C, D):
        (b, l, d_in) = u.shape
        n = A.shape[1]

        deltaA = torch.exp(einsum(delta, A, 'b l d_in, d_in n -> b l d_in n'))
        deltaB_u = einsum(delta, B, u, 'b l d_in, b l n, b l d_in -> b l d_in n')
        x = torch.zeros((b, d_in, n), device=deltaA.device)
        ys = []
        for i in range(l):
            x = deltaA[:, i] * x + deltaB_u[:, i]
            y = einsum(x, C[:, i, :], 'b d_in n, b n -> b d_in')
            ys.append(y)
        y = torch.stack(ys, dim=1)  # shape (b, l, d_in)

        y = y + u * D

        return y


class RMSNorm(nn.Module):
    def __init__(self,
                 d_model: int,
                 eps: float = 1e-5):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(d_model))

    def forward(self, x):
        output = x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps) * self.weight

        return output


def posemb_sincos_2d(h: int, w: int, dim: int, temperature: int = 10000) -> Tensor:
    y, x = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")

    if dim % 4 != 0:
        raise ValueError(f"feature dimension {dim} must be multiple of 4 for 2D sin-cos positional embedding")

    omega: Tensor = torch.arange(dim // 4) / (dim // 4 - 1)
    omega = 1.0 / (temperature ** omega)

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]

    pe = torch.cat((x.sin(), x.cos(), y.sin(), y.cos()), dim=1)

    return pe.to(torch.float32)


class MambaFFTClassifier(Module):
    def __init__(self, *,
                 input_shape: tuple[int, int, int],
                 n_classes: int,
                 n_patches: int,
                 d_model: int,
                 d_state: int,
                 d_conv: int,
                 expand: int,
                 n_layers: int,
                 layer_norm: bool = False,
                 conv_bias: bool = False,
                 bias: bool = False,
                 ) -> None:
        super().__init__()

        image_height: int = input_shape[0]
        image_width: int = input_shape[1]
        image_channels: int = input_shape[2]

        patch_height: int = image_height // n_patches
        patch_width: int = image_width // n_patches
        patch_size: int = patch_height * patch_width * image_channels

        freq_patch_size: int = patch_size * 2

        if image_height % n_patches != 0 or image_width % n_patches != 0:
            raise ValueError(f"image size ({input_shape[0]}, {input_shape[1]}) must be divisible by number of patches {n_patches}")

        self.to_patch_embedding = Sequential(
            Rearrange('b (h ph) (w pw) c -> b (h w) (ph pw c)', ph=patch_height, pw=patch_width),
            LayerNorm(patch_size) if layer_norm else Identity(),
            Linear(patch_size, d_model),
            LayerNorm(d_model) if layer_norm else Identity(),
        )

        self.to_freq_embedding = Sequential(
            Rearrange('b (h ph) (w pw) c ri -> b (h w) (ph pw c ri)', ph=patch_height, pw=patch_width),
            LayerNorm(freq_patch_size) if layer_norm else Identity(),
            Linear(freq_patch_size, d_model),
            LayerNorm(d_model) if layer_norm else Identity(),
        )

        self.pos_embedding = posemb_sincos_2d(
            h=image_height // patch_height,
            w=image_width // patch_width,
            dim=d_model,
        )

        self.layers: ModuleList = ModuleList([
            ResidualBlock(
                d_model=d_model,
                d_inner=d_model * expand,
                conv_bias=conv_bias,
                d_conv=d_conv,
                bias=bias,
                d_state=d_state,
                dt_rank=math.ceil(d_model / 16),
            )
            for _ in range(n_layers)
        ])

        self.norm_f = RMSNorm(d_model) if layer_norm else Identity()

        self.linear_head = Sequential(
            LayerNorm(d_model) if layer_norm else Identity(),
            Linear(d_model, n_classes),
        )

    def forward(self, input: Tensor) -> Tensor:
        freqs: Tensor = torch.view_as_real(fft2(input))

        x = self.to_patch_embedding(input)
        f = self.to_freq_embedding(freqs)

        x += self.pos_embedding.to(x.device)
        f += self.pos_embedding.to(f.device)

        x = torch.cat((x, f), dim=1)
        for layer in self.layers:
            x = layer(x)

        x = self.norm_f(x)
        x = x.mean(dim=1)

        x = self.linear_head(x)

        return x



from catasta import Scaffold, CatastaDataset
from catasta.models import MambaFFTRegressor
from catasta.transformations import (
    Normalization,
    WindowSliding,
    Slicing,
    Custom,
)
from catasta.dataclasses import EvalInfo


def main() -> None:
    n_dim: int = 768
    dataset_root: str = "data/nylon_wire/"
    input_trasnsformations = [
        Custom(lambda x: x[:10_000]),
        Normalization("minmax"),
        WindowSliding(window_size=n_dim, stride=1),
    ]
    output_trasnsformations = [
        Custom(lambda x: x[:10_000]),
        Normalization("minmax"),
        Slicing(amount=n_dim - 1, end="left"),
    ]

    dataset = CatastaDataset(
        root=dataset_root,
        task="regression",
        input_transformations=input_trasnsformations,
        output_transformations=output_trasnsformations,
    )

    model = MambaFFTRegressor(
        context_length=n_dim,
        n_patches=2,
        n_layers=2,
        d_model=8,
        d_conv=4,
        d_state=4,
        expand=1,
    )
    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="mse",
    )

    scaffold.train(
        epochs=10,
        batch_size=256,
        lr=1e-3,
        final_lr=1e-4,
        early_stopping=True,
    )

    info: EvalInfo = scaffold.evaluate()
    print(info)


if __name__ == '__main__':
    main()



from torch.nn import Module
from torch.nn.modules.loss import _Loss
from torch.nn import (
    MSELoss,
    L1Loss,
    SmoothL1Loss,
    HuberLoss,
    PoissonNLLLoss,
    KLDivLoss,
    CrossEntropyLoss,
    NLLLoss,
    BCELoss,
    BCEWithLogitsLoss,
    MarginRankingLoss,
    HingeEmbeddingLoss,
    MultiLabelMarginLoss,
    MultiLabelSoftMarginLoss,
)

from gpytorch.mlls import (
    VariationalELBO,
    PredictiveLogLikelihood,
    VariationalMarginalLogLikelihood,
    GammaRobustVariationalELBO,
)

available_loss_functions = sorted([
    "mse",
    "l1",
    "smooth_l1",
    "huber",
    "poisson",
    "kl_div",
    "cross_entropy",
    "nll",
    "bce",
    "bce_with_logits",
    "margin_ranking",
    "hinge_embedding",
    "multi_label_margin",
    "multi_label_soft_margin",
    "variational_elbo",
    "predictive_log",
    "variational_marginal_log",
    "gamma_robust_variational_elbo",
])

loss_functions: dict[str, _Loss] = {
    "mse": MSELoss(),
    "l1": L1Loss(),
    "smooth_l1": SmoothL1Loss(),
    "huber": HuberLoss(),
    "poisson": PoissonNLLLoss(),
    "kl_div": KLDivLoss(),
    "cross_entropy": CrossEntropyLoss(),
    "nll": NLLLoss(),
    "bce": BCELoss(),
    "bce_with_logits": BCEWithLogitsLoss(),
    "margin_ranking": MarginRankingLoss(),
    "hinge_embedding": HingeEmbeddingLoss(),
    "multi_label_margin": MultiLabelMarginLoss(),
    "multi_label_soft_margin": MultiLabelSoftMarginLoss(),
}


def get_loss_function(id: str | Module, model: Module, likelihood: Module, num_data: int) -> Module:
    if not isinstance(id, str):
        return id

    loss_function: _Loss | None = loss_functions.get(id.lower(), None)
    if loss_function is None:
        match id.lower():
            case "variational_elbo":
                return VariationalELBO(likelihood, model, num_data=num_data)
            case "predictive_log":
                return PredictiveLogLikelihood(likelihood, model, num_data=num_data)
            case "variational_marginal_log":
                return VariationalMarginalLogLikelihood(likelihood, model, num_data=num_data)
            case "gamma_robust_variational_elbo":
                return GammaRobustVariationalELBO(likelihood, model, num_data=num_data)
            case _:
                raise ValueError(f"invalid objective function id: {id}")

    return loss_function



from torch.nn import Module


class ModelStateManager:
    def __init__(self, early_stopping: bool) -> None:
        self.early_stopping = early_stopping

        self.best_loss: float = float("inf")
        self.best_model_states: list[dict] = []

        self.alpha: float = 0.95
        self.stop: bool = False

        self.prev_loss: float = float("inf")
        self.derivative: float = 0.0

    def __call__(self, models: list[Module], loss: float):
        if not self.best_model_states:
            self.best_model_states = [model.state_dict() for model in models]

        if loss < self.best_loss:
            self.best_loss = loss
            self.best_model_states = [model.state_dict() for model in models]

        unfiltered_derivate: float = loss - self.prev_loss if self.prev_loss != float("inf") else 0.0
        self.derivative = self.alpha * self.derivative + (1 - self.alpha) * unfiltered_derivate

        self.prev_loss = loss

        if self.early_stopping:
            if self.derivative > 0:
                self.stop = True

    def load_best_model_state(self, models: list[Module]) -> None:
        for i, model in enumerate(models):
            model.load_state_dict(self.best_model_states[i])



import numpy as np

from catasta.transformations import Decimation


def main() -> None:
    x: np.ndarray = np.arange(0, 100)
    decimation_factor: int = 2
    decimation: Decimation = Decimation(decimation_factor=decimation_factor)
    y: np.ndarray = decimation(x=x)

    assert np.allclose(y, np.arange(0, 100, decimation_factor))


if __name__ == "__main__":
    main()



import numpy as np

from torch.utils.data import DataLoader

from catasta.datasets import RegressionDataset
from catasta.transformations import (
    Custom,
    WindowSliding,
    Slicing,
)


def main() -> None:
    for n_dim in range(4, 13):
        input_transformations = [
            Custom(lambda x: x[7:904]),
            WindowSliding(window_size=n_dim, stride=1),
        ]
        output_transformations = [
            Custom(lambda x: x[7:904]),
            Slicing(amount=n_dim-1, end="left"),
        ]

        dataset = RegressionDataset(
            root="tests/data/test_data/",
            input_transformations=input_transformations,  # type: ignore
            output_transformations=output_transformations,  # type: ignore
            splits=(0.8, 0.1, 0.1),
        )

        train_dataloader = DataLoader(dataset.train, batch_size=3, shuffle=True)

        for x, y in train_dataloader:
            if x.mean() != y.mean():
                print("The values of x and y are not equal!")
                print(f"n_dim: {n_dim}")
                print(np.array(x), np.array(y))
                return

            if y.mean() >= 8:
                print("The value of y is too high!")
                print(f"n_dim: {n_dim}")
                print(np.array(x), np.array(y))
                return

        print("First test passed!")

        for x, y in dataset.validation:  # type: ignore
            if x.mean() != y.mean():
                print("The values of x and y are not equal!")
                print(f"n_dim: {n_dim}")
                print(np.array(x), np.array(y))
                return

            if y.mean() != 8:
                print("The value of y is not 8!")
                print(f"n_dim: {n_dim}")
                print(np.array(x), np.array(y))
                return

        print("Second test passed!")

        for x, y in dataset.test:  # type: ignore
            if x.mean() != y.mean():
                print("The values of x and y are not equal!")
                print(f"n_dim: {n_dim}")
                print(np.array(x), np.array(y))
                return

            if y.mean() != 9:
                print("The value of y is not 9!")
                print(f"n_dim: {n_dim}")
                print(np.array(x), np.array(y))
                return

        print("Third test passed!")


if __name__ == '__main__':
    main()



import platform
import time
import os

import numpy as np

import torch
from torch import Tensor
from torch.nn import Module, Identity
from torch.optim import Optimizer
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import DataLoader, Dataset
from torch.distributions import Distribution

from gpytorch.models.gp import GP
from gpytorch.mlls import MarginalLogLikelihood

from ..datasets import CatastaDataset
from .utils import (
    get_optimizer,
    get_loss_function,
    get_likelihood,
    ModelStateManager,
    TrainingLogger,
)
from ..dataclasses import TrainInfo, EvalInfo

from vclog import Logger


def _get_device(device: str) -> torch.device:
    match device:
        case "cpu":
            return torch.device("cpu")
        case "cuda":
            return torch.device("cuda")
        case "auto":
            return torch.device("cuda" if torch.cuda.is_available() else "cpu")
        case _:
            raise ValueError("Invalid device")


def _get_dtype(dtype: str) -> torch.dtype:
    match dtype:
        case "float16":
            return torch.float16
        case "float32":
            return torch.float32
        case "float64":
            return torch.float64
        case _:
            raise ValueError("Invalid dtype")


class Scaffold:
    def __init__(self, *,
                 model: Module,
                 dataset: CatastaDataset,
                 optimizer: str | Optimizer,
                 loss_function: str | Module,
                 probabilistic: bool = False,
                 likelihood: str | Module | None = None,
                 device: str = "auto",
                 dtype: str = "float32",
                 verbose: bool = True,
                 ) -> None:
        """The `Scaffold` class is a high-level API for training models on Catasta datasets.

        Arguments
        ---------
        model : torch.nn.Module
            The model to be trained.
        dataset : ~catasta.datasets.CatastaDataset
            The dataset to be used for training, validation, and testing.
        optimizer : str | torch.optim.Optimizer
            The optimizer to be used for training. The user can either pass a string with the name of the optimizer or a custom optimizer. One of: adam, sgd, adamw, lbfgs, rmsprop, rprop, adadelta, adagrad, adamax, asgd, sparseadam.
        loss_function : str | torch.nn.Module
            The loss function to be used for training. The user can either pass a string with the name of the loss function or a custom loss function. A list of available loss functions is in ~catasta.scaffolds.utils.available_loss_functions
        probabilistic : bool, optional
            Wheter to use probabilistic training or not. Defaults to False. Not yet supported.
        likelihood : str | torch.nn.Module, optional
            The likelihood to be used for Gaussian Processes. If the user is training a GP model and the likelihood is not provided, the default Gaussian likelihood is used. One of: gaussian, bernoulli, laplace, softmax, studentt, beta. 
        device : str, optional
            The device to be used for training. One of "cpu", "cuda", or "auto". Defaults to "auto".
        dtype : str, optional
            The data type to be used for training. One of "float16", "float32", or "float64". Defaults to "float32".
        verbose : bool, optional
            Whether to print training information or not. Defaults to True.
        """
        self.task: str = dataset.task

        self.device: torch.device = _get_device(device)
        self.dtype: torch.dtype = _get_dtype(dtype)

        self.model: Module = model.to(self.device, self.dtype)
        if isinstance(self.model, GP) and likelihood is None:
            self.likelihood: Module = get_likelihood("gaussian").to(self.device, self.dtype)
        else:
            self.likelihood: Module = get_likelihood(likelihood).to(self.device, self.dtype)

        self.dataset: CatastaDataset = dataset

        self.optimizer_id: str | Optimizer = optimizer
        self.loss_function_id: str | Module = loss_function

        self.logger: Logger = Logger("catasta", disable=not verbose)
        self._log_training_info()

        if probabilistic:
            self.logger.warning("probabilistic models are not yet supported")

    def train(self, *,
              epochs: int,
              batch_size: int,
              lr: float,
              final_lr: float | None = None,
              early_stopping: bool = False,
              shuffle: bool = True,
              ) -> TrainInfo:
        """Train the model on the dataset.

        Arguments
        ---------
        epochs : int
            The number of epochs to train the model.
        batch_size : int
            The batch size to use for training.
        lr : float
            The initial learning rate.
        final_lr : float, optional
            The final learning rate. If not provided, the learning rate is not decayed. Defaults to None.
        early_stopping : bool, optional
            Whether to use early stopping or not. The criterion for early stopping is the derivative of the validation loss. Defaults to False.
        shuffle : bool, optional
            Whether to shuffle the training data or not. Defaults to True.
        """
        # VARIABLES
        optimizer: Optimizer = get_optimizer(self.optimizer_id, [self.model, self.likelihood], lr)
        loss_function: Module = get_loss_function(self.loss_function_id, self.model, self.likelihood, len(self.dataset.train))  # type: ignore

        model_state_manager: ModelStateManager = ModelStateManager(early_stopping)
        lr_decay: float = (final_lr / lr) ** (1 / epochs) if final_lr else 1.0
        scheduler: StepLR = StepLR(optimizer, step_size=1, gamma=lr_decay)

        training_logger: TrainingLogger = TrainingLogger(self.task, epochs)

        train_data_loader: DataLoader = DataLoader(self.dataset.train, batch_size=batch_size, shuffle=shuffle)
        val_data_loader: DataLoader = DataLoader(self.dataset.validation, batch_size=batch_size, shuffle=False)

        # TRAINING LOOP
        for epoch in range(epochs):
            start_time: float = time.time()

            # train
            self.model.train()
            self.likelihood.train()

            train_loss, train_accuracy = _epoch(self.task, [self.model, self.likelihood], train_data_loader, optimizer, loss_function, self.device, self.dtype)

            # validation
            self.model.eval()
            self.likelihood.eval()

            with torch.no_grad():
                val_loss, val_accuracy = _epoch(self.task, [self.model, self.likelihood], val_data_loader, None, loss_function, self.device, self.dtype)

            scheduler.step()

            model_state_manager([self.model, self.likelihood], val_loss)

            if model_state_manager.stop:
                self.logger.warning(f"early stopping at epoch {epoch + 1}")
                break

            time_per_epoch = time.time() - start_time

            training_logger.log(
                train_loss=train_loss,
                train_accuracy=train_accuracy,
                val_loss=val_loss,
                val_accuracy=val_accuracy,
                lr=optimizer.param_groups[0]["lr"],
                epoch=epoch + 1,
                time_per_epoch=time_per_epoch,
            )

            Logger.plain(training_logger, color="green")

        # END OF TRAINING
        train_info = training_logger.get_info()

        model_state_manager.load_best_model_state([self.model, self.likelihood])

        return train_info

    def evaluate(self) -> EvalInfo:
        """Evaluate the model on the test set of the dataset.

        Returns
        -------
        ~catasta.dataclasses.EvalInfo
            The evaluation information.
        """
        self.model.eval()
        self.likelihood.eval()

        if self.task == "regression":
            return self._evaluate_regression(self.dataset.test)
        else:
            return self._evaluate_classification(self.dataset.test)

    def save(self,
             path: str,
             dtype: str = "float32",
             ) -> None:
        """Save the model to a file.

        Arguments
        ---------
        path : str
            The directory to save the model to.
        dtype : str, optional
            The data type to save the model in. Can be "float16", "float32", or "float64". Defaults to "float32".

        Raises
        ------
        ValueError
            If the path is a file path.
        """
        if "." in path:
            raise ValueError("save path must be a directory")

        model_dtype: torch.dtype = _get_dtype(dtype)
        model_device: torch.device = torch.device("cpu")

        save_path: str = os.path.join(path, self.model.__class__.__name__)

        os.makedirs(save_path, exist_ok=True)

        self.model.to(model_device, model_dtype)
        self.likelihood.to(model_device, model_dtype)

        for model in [self.model, self.likelihood]:
            if isinstance(model, Identity):
                continue

            model_path: str = os.path.join(save_path, f"{model.__class__.__name__}.pt")
            torch.save(model, model_path)

            self.logger.info(f"model saved to {model_path}")

    def _log_training_info(self) -> None:
        n_params: int = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        device_name: str = torch.cuda.get_device_name() if self.device.type == "cuda" else platform.processor()
        n_samples: int = len(self.dataset.train)  # type: ignore
        self.logger.info(f"""training
    -> task: {self.task}
    -> dataset: {self.dataset.root} ({n_samples} samples)
    -> model: {self.model.__class__.__name__} ({n_params} trainable parameters)
    -> device: {self.device} ({device_name}""")

    @torch.no_grad()
    def _evaluate_regression(self, dataset: Dataset) -> EvalInfo:
        x, y = next(iter(DataLoader(dataset, batch_size=len(dataset), shuffle=False)))  # type: ignore

        x: Tensor = x.to(self.device, dtype=self.dtype)
        y: Tensor = y.to(self.device, dtype=self.dtype)

        output = self.likelihood(self.model(x))

        true_output: np.ndarray = y.cpu().numpy()

        if isinstance(output, Distribution):
            predicted_output: np.ndarray = output.mean.cpu().numpy()
            predicted_output_std: np.ndarray = output.stddev.cpu().numpy()
        else:
            predicted_output: np.ndarray = output.cpu().numpy()
            predicted_output_std = np.zeros_like(true_output)

        return EvalInfo(
            task="regression",
            true_output=true_output,
            predicted_output=predicted_output,
            predicted_std=predicted_output_std,
        )

    @torch.no_grad()
    def _evaluate_classification(self, dataset: Dataset) -> EvalInfo:
        dataloader: DataLoader = DataLoader(dataset, batch_size=128, shuffle=False)

        true_labels: list[int] = []
        predicted_labels: list[int] = []
        for x_batch, y_batch in dataloader:
            x_batch = x_batch.to(self.device, dtype=self.dtype)
            y_batch = y_batch.to(self.device, dtype=torch.long)

            output: Tensor = self.model(x_batch)

            true_labels.append(y_batch.cpu().numpy())
            predicted_labels.append(output.argmax(dim=1).cpu().numpy())

        return EvalInfo(
            task="classification",
            true_output=np.concatenate(true_labels),
            predicted_output=np.concatenate(predicted_labels),
        )


def _epoch(task: str,
           models: list[Module],
           data_loader: DataLoader,
           optimizer: Optimizer | None,
           loss_function: Module,
           device: torch.device,
           dtype: torch.dtype,
           ) -> tuple[float, float]:
    cumulated_loss: float = 0.0
    total_samples: int = 0
    correct_predictions: int = 0
    for inputs, targets in data_loader:
        inputs: Tensor = inputs.to(device, dtype)
        targets: Tensor = targets.to(device, dtype if task == "regression" else torch.long)

        optimizer.zero_grad() if optimizer is not None else None

        outputs = models[0](inputs)
        if len(models) > 1:
            for model in models[1:]:
                outputs = model(outputs)

        loss: Tensor = loss_function(outputs, targets)  # type: ignore
        if isinstance(loss_function, MarginalLogLikelihood):
            loss = -loss

        loss.backward() if optimizer is not None else None
        optimizer.step() if optimizer is not None else None

        cumulated_loss += loss.item() * inputs.shape[0]
        total_samples += inputs.shape[0]
        if task == "classification":
            correct_predictions += (outputs.argmax(dim=1) == targets).sum().item()  # type: ignore

    epoch_loss = cumulated_loss / total_samples
    epoch_accuracy = correct_predictions / total_samples if task == "classification" else -np.inf

    return epoch_loss, epoch_accuracy



from torch.nn import Module
from torch.optim import (
    Optimizer,
    Adam,
    SGD,
    AdamW,
    LBFGS,
    RMSprop,
    Rprop,
    Adadelta,
    Adagrad,
    Adamax,
    ASGD,
    SparseAdam,
)


def get_optimizer(id: str | Optimizer, model: list[Module], lr: float) -> Optimizer:
    if isinstance(id, Optimizer):
        return id

    parameters = []
    for m in model:
        parameters += list(m.parameters())

    match id.lower():
        case "adam":
            return Adam(parameters, lr=lr)
        case "sgd":
            return SGD(parameters, lr=lr)
        case "adamw":
            return AdamW(parameters, lr=lr)
        case "lbfgs":
            return LBFGS(parameters, lr=lr)
        case "rmsprop":
            return RMSprop(parameters, lr=lr)
        case "rprop":
            return Rprop(parameters, lr=lr)
        case "adadelta":
            return Adadelta(parameters, lr=lr)
        case "adagrad":
            return Adagrad(parameters, lr=lr)
        case "adamax":
            return Adamax(parameters, lr=lr)
        case "asgd":
            return ASGD(parameters, lr=lr)
        case "sparseadam":
            return SparseAdam(parameters, lr=lr)
        case _:
            raise ValueError(f"invalid optimizer id: {id}")



import pandas as pd
import matplotlib.pyplot as plt

from catasta import Scaffold, CatastaDataset, Archway
from catasta.models import ApproximateGPRegressor
from catasta.scaffolds import Scaffold
from catasta.dataclasses import PredictionInfo, TrainInfo


def main() -> None:
    # Model
    model = ApproximateGPRegressor(
        n_inducing_points=128,
        context_length=1,
        kernel="matern",
        mean="constant",
    )

    # Dataset
    dataset_root: str = "data/incomplete/"
    dataset = CatastaDataset(dataset_root, task="regression")

    # Training
    scaffold = Scaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="variational_elbo",
    )

    train_info: TrainInfo = scaffold.train(
        epochs=5000,
        batch_size=128,
        lr=1e-3,
        early_stopping=True,
    )
    print(f"min train loss: {train_info.best_train_loss:.4f}")

    info = scaffold.evaluate()
    print(info)

    # save model
    save_model_path: str = "saved_models/"
    scaffold.save(save_model_path)

    # Inference
    df = pd.read_csv(dataset_root + "data.csv")
    observed_input = df["input"].to_numpy()
    observed_output = df["output"].to_numpy()
    true_input = df["true_input"].to_numpy()
    true_output = df["true_output"].to_numpy()

    archway = Archway(
        path=save_model_path+model.__class__.__name__,
        verbose=False,
    )
    predictions: PredictionInfo = archway.predict(true_input.reshape(-1, 1))

    predicted_output = predictions.value
    predicted_std = predictions.std

    # Plot
    plt.plot(true_input, true_output, color="red", label="true")
    plt.plot(observed_input, observed_output, "k.", label="observations")
    plt.plot(true_input, predicted_output, color="blue", label="predicted")
    plt.fill_between(
        true_input,
        predicted_output - predicted_std,
        predicted_output + predicted_std,
        color="blue",
        alpha=0.2,
    )
    plt.show()


if __name__ == '__main__':
    main()



from catasta.models import CNNClassifier
from catasta.datasets import ClassificationDataset
from catasta.scaffolds import ImageClassificationScaffold
from catasta.dataclasses import ClassificationEvalInfo

from vclog import Logger


def main() -> None:
    logger: Logger = Logger("catasta")

    model = CNNClassifier(
        input_shape=(28, 28, 3),
        n_classes=10,
        conv_out_channels=[32, 64],
        conv_kernel_sizes=[3, 3],
        conv_strides=[1, 1],
        conv_paddings=[1, 1],
        pooling_kernel_sizes=[2, 2],
        pooling_strides=[2, 2],
        pooling_paddings=[0, 0],
        feedforward_dims=[128, 64],
        dropout=0.5,
        activation="relu",
    )

    dataset = ClassificationDataset(
        root="tests/data/mnist",
    )

    scaffold = ImageClassificationScaffold(
        model=model,
        dataset=dataset,
        optimizer="adamw",
        loss_function="cross_entropy",
    )

    scaffold.train(
        epochs=10,
        batch_size=128,
        lr=1e-3,
        final_lr=None,
        early_stopping=None,
    )

    eval_info: ClassificationEvalInfo = scaffold.evaluate()
    logger.info(eval_info)

    scaffold.save(path="tests/models/", to_onnx=True)


if __name__ == '__main__':
    main()



import torch
from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Sequential,
    ModuleList,
    Dropout,
    Softmax,
    GELU,
    Identity,
)
from torch.fft import fft2

from einops import rearrange
from einops.layers.torch import Rearrange


def posemb_sincos_2d(h: int, w: int, dim: int, temperature: int = 10000) -> Tensor:
    y, x = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")

    if dim % 4 != 0:
        raise ValueError(f"feature dimension {dim} must be multiple of 4 for 2D sin-cos positional embedding")

    omega: Tensor = torch.arange(dim // 4) / (dim // 4 - 1)
    omega = 1.0 / (temperature ** omega)

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]

    pe = torch.cat((x.sin(), x.cos(), y.sin(), y.cos()), dim=1)

    return pe.to(torch.float32)


class FeedForward(Module):
    def __init__(self, *,
                 d_model: int,
                 hidden_dim: int,
                 dropout: float,
                 layer_norm: bool,
                 ) -> None:
        super().__init__()

        self.net = Sequential(
            LayerNorm(d_model) if layer_norm else Identity(),
            Linear(d_model, hidden_dim),
            GELU(),
            Dropout(dropout),
            Linear(hidden_dim, d_model),
            Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.net(x)


class Attention(Module):
    def __init__(self, *,
                 d_model: int,
                 n_heads: int,
                 head_dim: int,
                 dropout: float,
                 layer_norm: bool,
                 ) -> None:
        super().__init__()

        self.n_heads: int = n_heads
        self.scale: int = head_dim ** -0.5
        self.norm = LayerNorm(d_model) if layer_norm else Identity()
        self.attend = Softmax(dim=-1)
        self.dropout = Dropout(dropout)

        inner_dim: int = head_dim * n_heads
        self.to_qkv = Linear(d_model, inner_dim * 3, bias=False)
        self.to_out = Sequential(
            Linear(inner_dim, d_model, bias=False),
            Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.norm(x)

        qkv: Tensor = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.n_heads), qkv)

        dots: Tensor = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn: Tensor = self.attend(dots)
        attn = self.dropout(attn)

        out: Tensor = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')

        return self.to_out(out)


class Transformer(Module):
    def __init__(self, *,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 head_dim: int,
                 feedforward_dim: int,
                 dropout: float,
                 layer_norm: bool,
                 ) -> None:
        super().__init__()

        self.norm = LayerNorm(d_model) if layer_norm else Identity()
        self.layers = ModuleList([])
        for _ in range(n_layers):
            self.layers.append(ModuleList([
                Attention(d_model=d_model,
                          n_heads=n_heads,
                          head_dim=head_dim,
                          dropout=dropout,
                          layer_norm=layer_norm,
                          ),
                FeedForward(d_model=d_model,
                            hidden_dim=feedforward_dim,
                            dropout=dropout,
                            layer_norm=layer_norm,
                            )
            ]))

    def forward(self, x: Tensor) -> Tensor:
        for attn, ff in self.layers:  # type: ignore
            x = attn(x) + x
            x = ff(x) + x

        x = self.norm(x)

        return x


class TransformerFFTClassifier(Module):
    def __init__(self, *,
                 input_shape: tuple[int, int, int],
                 n_classes: int,
                 n_patches: int,
                 d_model: int,
                 n_layers: int,
                 n_heads: int,
                 feedforward_dim: int,
                 head_dim: int,
                 dropout: float = 0.0,
                 layer_norm: bool = False,
                 ) -> None:
        super().__init__()

        image_height: int = input_shape[0]
        image_width: int = input_shape[1]
        image_channels: int = input_shape[2]

        patch_height: int = image_height // n_patches
        patch_width: int = image_width // n_patches
        patch_size: int = patch_height * patch_width * image_channels

        freq_patch_size: int = patch_size * 2

        if image_height % n_patches != 0 or image_width % n_patches != 0:
            raise ValueError(f"image size ({input_shape[0]}, {input_shape[1]}) must be divisible by number of patches {n_patches}")

        self.to_patch_embedding = Sequential(
            Rearrange('b (h ph) (w pw) c -> b (h w) (ph pw c)', ph=patch_height, pw=patch_width),
            LayerNorm(patch_size) if layer_norm else Identity(),
            Linear(patch_size, d_model),
            LayerNorm(d_model) if layer_norm else Identity(),
        )

        self.to_freq_embedding = Sequential(
            Rearrange('b (h ph) (w pw) c ri -> b (h w) (ph pw c ri)', ph=patch_height, pw=patch_width),
            LayerNorm(freq_patch_size) if layer_norm else Identity(),
            Linear(freq_patch_size, d_model),
            LayerNorm(d_model) if layer_norm else Identity(),
        )

        self.pos_embedding = posemb_sincos_2d(
            h=image_height // patch_height,
            w=image_width // patch_width,
            dim=d_model,
        )
        self.dropout = Dropout(dropout)

        self.transformer = Transformer(
            d_model=d_model,
            n_layers=n_layers,
            n_heads=n_heads,
            head_dim=head_dim,
            feedforward_dim=feedforward_dim,
            dropout=dropout,
            layer_norm=layer_norm,
        )

        self.linear_head = Sequential(
            LayerNorm(d_model) if layer_norm else Identity(),
            Linear(d_model, n_classes),
        )

    def forward(self, input: Tensor) -> Tensor:
        freqs: Tensor = torch.view_as_real(fft2(input))

        x = self.to_patch_embedding(input)
        f = self.to_freq_embedding(freqs)

        x += self.pos_embedding.to(x.device)
        f += self.pos_embedding.to(f.device)

        x = torch.cat((x, f), dim=1)

        x = self.transformer(x)
        x = x.mean(dim=1)

        x = self.linear_head(x)

        return x



import torch
from torch import Tensor
from torch.nn import (
    Conv2d,
    Module,
    ModuleList,
    MaxPool2d,
    Linear,
    Flatten,
    Dropout,
    ReLU,
    Sigmoid,
    Tanh,
    GELU,
)

from einops import rearrange


def get_activation_function(activation: str) -> Module:
    match activation:
        case 'relu':
            return ReLU()
        case 'sigmoid':
            return Sigmoid()
        case 'tanh':
            return Tanh()
        case 'gelu':
            return GELU()
        case _:
            raise ValueError(f'Activation function {activation} not supported')


class CNNClassifier(Module):
    def __init__(self, *,
                 input_shape: tuple[int, int, int],
                 n_classes: int,
                 conv_out_channels: list[int],
                 feedforward_dims: list[int],
                 conv_kernel_size: int = 3,
                 conv_stride: int = 1,
                 conv_padding: int = 0,
                 pooling_kernel_size: int = 2,
                 pooling_stride: int = 2,
                 pooling_padding: int = 0,
                 dropout: float = 0.5,
                 activation: str = 'relu',
                 ) -> None:
        super().__init__()
        conv_kernel_sizes = [conv_kernel_size] * len(conv_out_channels)
        conv_strides = [conv_stride] * len(conv_out_channels)
        conv_paddings = [conv_padding] * len(conv_out_channels)
        pooling_kernel_sizes = [pooling_kernel_size] * len(conv_out_channels)
        pooling_strides = [pooling_stride] * len(conv_out_channels)
        pooling_paddings = [pooling_padding] * len(conv_out_channels)

        n_input_channels: int = input_shape[-1]

        # CONV LAYERS
        self.conv_layers = ModuleList()
        for i in range(len(conv_out_channels)):
            conv2d = Conv2d(
                in_channels=n_input_channels if i == 0 else conv_out_channels[i - 1],
                out_channels=conv_out_channels[i],
                kernel_size=conv_kernel_sizes[i],
                stride=conv_strides[i],
                padding=conv_paddings[i],
            )

            activation_function: Module = get_activation_function(activation)

            max_pool2d = MaxPool2d(
                kernel_size=pooling_kernel_sizes[i],
                stride=pooling_strides[i],
                padding=pooling_paddings[i],
            )

            self.conv_layers.append(conv2d)
            self.conv_layers.append(activation_function)
            self.conv_layers.append(max_pool2d)

        # FLATTEN
        self.flatten = Flatten()

        # PREVIOUS NEURONS
        dummy_input = torch.randn(input_shape)
        dummy_input = rearrange(dummy_input, 'h w c -> 1 c h w')

        for layer in self.conv_layers:
            dummy_input = layer(dummy_input)

        dummy_input = self.flatten(dummy_input)
        prev_neurons = dummy_input.shape[1]

        # DENSE LAYERS
        self.dense_layers = ModuleList()
        for i in range(len(feedforward_dims)):
            linear = Linear(
                in_features=prev_neurons if i == 0 else feedforward_dims[i - 1],
                out_features=feedforward_dims[i],
            )

            activation_function: Module = get_activation_function(activation)
            dropout_layer = Dropout(p=dropout)

            self.dense_layers.append(linear)
            self.dense_layers.append(activation_function)
            self.dense_layers.append(dropout_layer)

        linear = Linear(
            in_features=feedforward_dims[-1],
            out_features=n_classes,
        )

        self.dense_layers.append(linear)

    def forward(self, x: Tensor) -> Tensor:
        x = rearrange(x, 'b h w c -> b c h w')

        for layer in self.conv_layers:
            x = layer(x)

        x = self.flatten(x)

        for layer in self.dense_layers:
            x = layer(x)

        return x



import torch
from torch import Tensor

from gpytorch.models import ApproximateGP
from gpytorch.variational import CholeskyVariationalDistribution, VariationalStrategy
from gpytorch.means import ZeroMean, ConstantMean, Mean
from gpytorch.kernels import Kernel, ScaleKernel, RBFKernel, MaternKernel, RQKernel, RFFKernel, PeriodicKernel, LinearKernel
from gpytorch.distributions import MultivariateNormal


def _get_kernel(id: str, context_length: int, use_ard: bool) -> Kernel:
    match id.lower():
        case "rq":
            return RQKernel(ard_num_dims=context_length if use_ard else None)
        case "matern":
            return MaternKernel(ard_num_dims=context_length if use_ard else None)
        case "rbf":
            return RBFKernel(ard_num_dims=context_length if use_ard else None)
        case "rff":
            return RFFKernel(num_samples=context_length)
        case "periodic":
            return PeriodicKernel(ard_num_dims=context_length if use_ard else None)
        case "linear":
            return LinearKernel(ard_num_dims=context_length if use_ard else None)
        case _:
            raise ValueError(f"Unknown kernel: {id}")


def _get_mean_module(id: str) -> Mean:
    match id.lower():
        case "constant":
            return ConstantMean()
        case "zero":
            return ZeroMean()
        case _:
            raise ValueError(f"Unknown mean: {id}")


class ApproximateGPRegressor(ApproximateGP):
    def __init__(self, *,
                 n_inducing_points: int,
                 context_length: int,
                 kernel: str = "rq",
                 mean: str = "constant",
                 use_ard: bool = False,
                 ) -> None:
        '''
        Arguments
        ---------
        n_inducing_points: int
            Number of inducing points
        context_length: int
            Number of input dimensions
        kernel: str
            Kernel to use. One of "rq", "matern", "rbf", "rff", "periodic", "linear"
        mean: str
            Mean to use. One of "constant", "zero"
        use_ard: bool
            Whether to use Automatic Relevance Determination
        '''
        inducing_points: torch.Tensor = torch.randn(n_inducing_points, context_length, dtype=torch.float32)

        variational_distribution = CholeskyVariationalDistribution(n_inducing_points)
        variational_strategy = VariationalStrategy(self, inducing_points, variational_distribution)

        super().__init__(variational_strategy)
        self.mean_module = _get_mean_module(mean)
        self.covar_module = ScaleKernel(_get_kernel(kernel, context_length, use_ard))

    def forward(self, x: Tensor) -> MultivariateNormal:
        mean_x: Tensor = self.mean_module(x)  # type: ignore
        covar_x: Tensor = self.covar_module(x)  # type: ignore

        return MultivariateNormal(mean_x, covar_x)



from catasta.models import TransformerSignalClassifier
from catasta.scaffolds import ClassifierScaffold
from catasta.datasets import SignalDataset
from catasta.entities import TrainInfo

import matplotlib.pyplot as plt


def main() -> None:
    # Model parameters
    sequence_length: int = 256
    n_heads: int = 4
    n_embeddings: int = 256
    n_layers: int = 6
    feedforward_dim: int = 512

    # Training parameters
    batch_size: int = 8
    learning_rate: float = 6e-4
    epochs: int = 20
    log_interval: int = 1
    eval_iters: int = 10

    dataset = SignalDataset(root="tests/signals", sequence_length=sequence_length)

    model = TransformerSignalClassifier(embedding_dim=n_embeddings,
                                        patch_size=sequence_length//32,
                                        n_heads=n_heads,
                                        n_layers=n_layers,
                                        sequence_length=sequence_length,
                                        feedforward_dim=feedforward_dim,
                                        n_classes=dataset.n_classes,
                                        )

    scaffold = ClassifierScaffold(model=model, dataset=dataset)

    # Train model
    train_info_list: list[TrainInfo] = scaffold.train(batch_size=batch_size,
                                                      learning_rate=learning_rate,
                                                      epochs=epochs,
                                                      log_interval=log_interval,
                                                      eval_iters=eval_iters,
                                                      )

    # Plot training info
    train_loss_list: list[float] = [train_info.train_loss for train_info in train_info_list]
    val_loss_list: list[float] = [train_info.val_loss for train_info in train_info_list]
    train_accuracy_list: list[float] = [train_info.train_acc for train_info in train_info_list]
    val_accuracy_list: list[float] = [train_info.val_acc for train_info in train_info_list]

    plt.figure()
    plt.plot(train_loss_list, label='Train loss')
    plt.plot(val_loss_list, label='Validation loss')
    plt.legend()

    plt.figure()
    plt.plot(train_accuracy_list, label='Train accuracy')
    plt.plot(val_accuracy_list, label='Validation accuracy')
    plt.legend()

    plt.show()


if __name__ == "__main__":
    main()



import os
import numpy as np
from PIL import Image

from catasta.models import CNNClassifier
from catasta.archways import ClassificationArchway
from catasta.dataclasses import ClassificationEvalInfo


def vanilla() -> None:
    path: str = "tests/models/"

    model = CNNClassifier(
        input_shape=(28, 28, 3),
        n_classes=10,
        conv_out_channels=[32, 64],
        conv_kernel_sizes=[3, 3],
        conv_strides=[1, 1],
        conv_paddings=[1, 1],
        pooling_kernel_sizes=[2, 2],
        pooling_strides=[2, 2],
        pooling_paddings=[0, 0],
        feedforward_dims=[128, 64],
        dropout=0.5,
        activation="relu",
    )

    archway = ClassificationArchway(
        model=model,
        path=path,
    )

    imgs = []
    dataset_root = "tests/data/mnist/validation/0"
    for file in os.listdir(dataset_root):
        img = Image.open(os.path.join(dataset_root, file))
        img = np.array(img.convert("RGB"))
        imgs.append(img)

    input = np.array(imgs)
    prediction = archway.predict(input)

    predicted_output = prediction.argmax
    true_labels = np.zeros(len(predicted_output)).astype(int)

    info = ClassificationEvalInfo(
        true_labels=true_labels,
        predicted_labels=predicted_output,
        n_classes=10,
    )

    print(info)


def vanilla_onnx() -> None:
    path: str = "tests/models/"

    archway = ClassificationArchway(
        path=path,
        from_onnx=True,
    )

    imgs = []
    dataset_root = "tests/data/mnist/validation/0"
    for file in os.listdir(dataset_root):
        img = Image.open(os.path.join(dataset_root, file))
        img = np.array(img.convert("RGB"))
        imgs.append(img)

    input = np.array(imgs)
    prediction = archway.predict(input)

    predicted_output = prediction.argmax
    true_labels = np.zeros(len(predicted_output)).astype(int)

    info = ClassificationEvalInfo(
        true_labels=true_labels,
        predicted_labels=predicted_output,
        n_classes=10,
    )

    print(info)


if __name__ == "__main__":
    print("vanilla")
    vanilla()
    print("vanilla_onnx")
    vanilla_onnx()



import os
import shutil

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from catasta.entities import ModelData
from catasta.scaffolds import PredictorScaffold

# from gluonts.torch.model.deepar import DeepAREstimator
from gluonts.torch.model.patch_tst import PatchTSTEstimator


def get_data(dir: str) -> list[ModelData]:
    data: list[ModelData] = []
    for filename in os.listdir(dir):
        if not filename.endswith(".csv"):
            continue

        data_frame: pd.DataFrame = pd.read_csv(dir + filename)
        data.append(ModelData(
            input=data_frame['input'].to_numpy(),
            output=data_frame['output'].to_numpy(),
        ))

    return data


def escalonate(data: ModelData) -> list[ModelData]:
    return [ModelData(
        input=data.input[:i],
        output=data.output[:i],
    ) for i in range(1, len(data.input))]


def main() -> None:
    # data: list[ModelData] = get_data("data/steps/")
    data: list[ModelData] = get_data("tests/data/nylon_strain/")

    # Train
    prediction_length: int = 12
    epochs: int = 100
    context: int = int(len(data[0].input)*0.15)

    # model = DeepAREstimator(
    #     prediction_length=prediction_length,
    #     num_layers=4,
    #     hidden_size=64,
    #     freq="D",
    #     num_feat_dynamic_real=1,
    # )

    model = PatchTSTEstimator(
        prediction_length=prediction_length,
        patch_len=prediction_length,
        d_model=32,
        nhead=4,
        dim_feedforward=128,
    )

    scaffold = PredictorScaffold(model)

    scaffold.train(dataset=data[:-1],
                   epochs=epochs,
                   lr=0.001,
                   train_ratio=0.8,
                   )

    test_data: list[ModelData] = escalonate(data[-1])[context:]
    inputs: list[np.ndarray] = [d.input for d in test_data]
    outputs: list[np.ndarray] = [d.output for d in test_data]

    # WITH REAL OUTPUT
    predicted_output: np.ndarray = outputs[0]
    q: np.ndarray = outputs[0][-prediction_length:]
    for input, output in zip(inputs, outputs):
        assert len(q) == prediction_length
        assert len(input) == len(output)

        d = ModelData(input, output)
        forecasts, _ = scaffold.predict(d)

        # p_hat = forecasts[0].mean
        # overlapping_mean: np.ndarray = np.mean([q[1:], p_hat[:-1]], axis=0)
        # q = np.append(overlapping_mean, p_hat[-1])
        # predicted_output = np.append(predicted_output, q[0])

        predicted_output = np.append(predicted_output, np.mean(forecasts[0].mean))

    plt.figure(figsize=(30, 20))
    plt.plot(data[-1].output, label="Real")
    plt.plot(predicted_output, label="Predicted")
    plt.legend()
    plt.show()
    plt.close()

    # WITHOUT REAL OUTPUT
    predicted_output: np.ndarray = outputs[0]
    q: np.ndarray = outputs[0][-prediction_length:]
    for input in inputs:
        assert len(q) == prediction_length
        assert len(input) == len(predicted_output)

        d = ModelData(input, predicted_output)

        forecasts, _ = scaffold.predict(d)

        # p_hat = forecasts[0].mean
        # overlapping_mean: np.ndarray = np.mean([q[1:], p_hat[:-1]], axis=0)
        # q = np.append(overlapping_mean, p_hat[-1])
        # predicted_output = np.append(predicted_output, q[0])
        predicted_output = np.append(predicted_output, np.mean(forecasts[0].mean))

    plt.figure(figsize=(30, 20))
    plt.plot(data[-1].output, label="Real")
    plt.plot(predicted_output, label="Predicted")
    plt.plot(data[-1].output[:context], color="green")
    plt.legend()
    plt.show()
    plt.close()


if __name__ == "__main__":
    main()
    if os.path.exists("lightning_logs"):
        shutil.rmtree("lightning_logs")



import numpy as np

from .transformation import Transformation

import numpy as np


class KalmanFilter(Transformation):
    def __init__(self, *,
                 F: np.ndarray,
                 H: np.ndarray,
                 B: int | None = None,
                 Q: np.ndarray | None = None,
                 R: np.ndarray | None = None,
                 P: np.ndarray | None = None,
                 x0: np.ndarray | None = None,
                 ) -> None:
        self.n: int = F.shape[1]
        self.m: int = H.shape[1]

        self.F: np.ndarray = F
        self.H: np.ndarray = H
        self.B: int = 0 if B is None else B
        self.Q: np.ndarray = np.eye(self.n) if Q is None else Q
        self.R: np.ndarray = np.eye(self.n) if R is None else R
        self.P: np.ndarray = np.eye(self.n) if P is None else P
        self.x: np.ndarray = np.zeros((self.n, 1)) if x0 is None else x0

    def predict(self, u: int = 0) -> np.ndarray:
        self.x = np.dot(self.F, self.x) + np.dot(self.B, u)
        self.P = np.dot(np.dot(self.F, self.P), self.F.T) + self.Q

        return self.x

    def update(self, z: int) -> None:
        y = z - np.dot(self.H, self.x)

        S = self.R + np.dot(self.H, np.dot(self.P, self.H.T))
        K = np.dot(np.dot(self.P, self.H.T), np.linalg.inv(S))

        self.x = self.x + np.dot(K, y)

        I = np.eye(self.n)

        self.P = np.dot(np.dot(I - np.dot(K, self.H), self.P), (I - np.dot(K, self.H)).T) + np.dot(np.dot(K, self.R), K.T)

    def __call__(self, x: np.ndarray) -> np.ndarray:
        predictions: list[np.ndarray] = []
        for z in x:
            predictions.append(np.dot(self.H, self.predict())[0])
            self.update(z)

        return np.array(predictions).flatten()



from typing import NamedTuple
import numpy as np


class PredictionInfo(NamedTuple):
    """NamedTuple for storing prediction information.

    Attributes
    ----------
    value : np.ndarray
        The predicted values.
    argmax : np.ndarray
        The argmax of the predicted values. If the task is regression, this array will be empty.
    std : np.ndarray
        The standard deviation of the predicted values. If the task is regression, this array will be a vector of zeroes.
    """
    value: np.ndarray
    std: np.ndarray
    argmax: np.ndarray



import numpy as np
import matplotlib.pyplot as plt

from catasta.transformations import KalmanFilter


def main() -> None:
    noisy_sin = np.sin(np.linspace(0, 100, 1000)) + np.random.normal(0, 0.1, 1000)

    kf = KalmanFilter(
        F=np.array([[1, 1], [0, 1]]),
        H=np.array([[1, 0]]),
        Q=np.array([[0.0001, 0], [0, 0.0001]]),
        R=np.array([[1]]),
    )

    predictions = kf(noisy_sin)

    plt.figure(figsize=(30, 20))
    plt.plot(noisy_sin, label="noisy sin", color="red")
    plt.plot(predictions, label="kalman filter", color="blue")
    plt.legend()
    plt.show()


if __name__ == "__main__":
    main()



import numpy as np


class EvalInfo:
    """A class that calculates and stores evaluation metrics for regression and classification tasks.

    Attributes
    ----------
    self.task : str
        The task type, either 'regression' or 'classification'.
    self.true_output : np.ndarray
        The true output values.
    self.predicted_output : np.ndarray
        The predicted output values.
    self.predicted_std : np.ndarray
        The standard deviation of the predicted output values.
    self.mae : float
        The mean absolute error. Only available for regression tasks.
    self.mse : float
        The mean squared error. Only available for regression tasks.
    self.rmse : float
        The root mean squared error. Only available for regression tasks.
    self.r2 : float
        The R2 score. Only available for regression tasks.
    self.confusion_matrix : np.ndarray
        The confusion matrix. Only available for classification tasks.
    self.tp : np.ndarray
        The true positive values. Only available for classification tasks.
    self.fp : np.ndarray
        The false positive values. Only available for classification tasks.
    self.fn : np.ndarray
        The false negative values. Only available for classification tasks.
    self.tn : np.ndarray
        The true negative values. Only available for classification tasks.
    self.accuracy : float
        The accuracy score. Only available for classification tasks.
    self.precision : float
        The precision score. Only available for classification tasks.
    self.recall : float
        The recall score. Only available for classification tasks.
    self.sensitivity : float
        The sensitivity score. Only available for classification tasks.
    self.specificity : float
        The specificity score. Only available for classification tasks.
    self.f1_score : float
        The F1 score. Only available for classification tasks.
    """

    def __init__(self,
                 task: str,
                 true_output: np.ndarray,
                 predicted_output: np.ndarray,
                 predicted_std: np.ndarray | None = None,
                 n_classes: int | None = None
                 ) -> None:
        """
        Parameters
        ----------
        task : str
            The task type, either 'regression' or 'classification'.
        true_output : np.ndarray
            The true output values.
        predicted_output : np.ndarray
            The predicted output values.
        predicted_std : np.ndarray, optional
            The standard deviation of the predicted output values. If not provided, it is set to an array of zeros.
        n_classes : int, optional
            The number of classes in the classification task. If not provided, it is set to the number of unique classes.
        """

        self.task = task
        self.true_output = true_output
        self.predicted_output = predicted_output
        self.predicted_std = predicted_std if predicted_std is not None else np.zeros_like(predicted_output)

        if self.task == 'regression':
            self._calculate_regression_metrics(self.true_output, self.predicted_output)
        elif self.task == 'classification':
            self._calculate_classification_metrics(self.true_output, self.predicted_output, n_classes)
        else:
            raise ValueError('Invalid task type')

    def _calculate_regression_metrics(self, y: np.ndarray, y_pred: np.ndarray) -> None:
        self.mae: float = np.mean(np.abs(y - y_pred)).astype(float)
        self.mse: float = np.mean((y - y_pred)**2).astype(float)
        self.rmse: float = np.sqrt(np.mean((y - y_pred)**2))
        self.r2: float = 1 - np.sum((y - y_pred)**2) / np.sum((y - np.mean(y))**2).astype(float)

    def _calculate_classification_metrics(self, y: np.ndarray, y_pred: np.ndarray, n_classes: int | None) -> None:
        self.confusion_matrix: np.ndarray = self._compute_confusion_matrix(y, y_pred, n_classes)

        self.tp: np.ndarray = np.diag(self.confusion_matrix)
        self.fp: np.ndarray = np.sum(self.confusion_matrix, axis=0) - self.tp
        self.fn: np.ndarray = np.sum(self.confusion_matrix, axis=1) - self.tp
        self.tn: np.ndarray = np.sum(self.confusion_matrix) - (self.tp + self.fp + self.fn)

        tp: int = np.sum(self.tp)
        fp: int = np.sum(self.fp)
        fn: int = np.sum(self.fn)
        tn: int = np.sum(self.tn)

        self.accuracy: float = (tp + tn) / (tp + fp + fn + tn)
        self.precision: float = tp / (tp + fp)
        self.recall = self.sensitivity = tp / (tp + fn)
        self.specificity: float = tn / (tn + fp)
        self.f1_score: float = 2 * (self.precision * self.recall) / (self.precision + self.recall)

    def _compute_confusion_matrix(self, y: np.ndarray, y_pred: np.ndarray, n_classes: int | None) -> np.ndarray:
        classes: int = len(np.unique(y)) if n_classes is None else n_classes

        matrix: np.ndarray = np.zeros((classes, classes), dtype=int)
        for true, pred in zip(y, y_pred):
            matrix[true, pred] += 1

        return matrix

    def _regression_repr(self) -> str:
        mae: str = f'MAE: {self.mae:.4f}\n'
        mse: str = f'MSE: {self.mse:.4f}\n'
        rmse: str = f'RMSE: {self.rmse:.4f}\n'
        r2: str = f'R2: {self.r2:.4f}'

        return mae + mse + rmse + r2

    def _classification_repr(self) -> str:
        accuracy: str = f'Accuracy: {self.accuracy:.4f}\n'
        precision: str = f'Precision: {self.precision:.4f}\n'
        recall: str = f'Recall: {self.recall:.4f}\n'
        sensitivity: str = f'Sensitivity: {self.sensitivity:.4f}\n'
        specificity: str = f'Specificity: {self.specificity:.4f}\n'
        f1_score: str = f'F1 Score: {self.f1_score:.4f}'

        return accuracy + precision + recall + sensitivity + specificity + f1_score

    def __repr__(self) -> str:
        if self.task == 'regression':
            return self._regression_repr()
        elif self.task == 'classification':
            return self._classification_repr()
        else:
            raise ValueError('Invalid task type')



import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat, einsum

from torch import Tensor
from torch.nn import (
    Module,
    Linear,
    LayerNorm,
    Sequential,
    ModuleList,
    Identity,
)

from einops.layers.torch import Rearrange


class ResidualBlock(Module):
    def __init__(self, *,
                 d_model: int,
                 d_inner: int,
                 conv_bias: bool,
                 d_conv: int,
                 bias: bool,
                 d_state: int,
                 dt_rank: int,
                 ) -> None:
        super().__init__()
        self.mixer = MambaBlock(
            d_model=d_model,
            d_inner=d_inner,
            conv_bias=conv_bias,
            d_conv=d_conv,
            bias=bias,
            d_state=d_state,
            dt_rank=dt_rank,
        )
        self.norm = RMSNorm(d_model)

    def forward(self, x):
        output = self.mixer(self.norm(x)) + x

        return output


class MambaBlock(Module):
    def __init__(self, *,
                 d_model: int,
                 d_inner: int,
                 conv_bias: bool,
                 d_conv: int,
                 bias: bool,
                 d_state: int,
                 dt_rank: int,
                 ) -> None:
        super().__init__()

        self.d_inner = d_inner
        self.dt_rank = dt_rank

        self.in_proj = nn.Linear(d_model, d_inner * 2, bias=bias)

        self.conv1d = nn.Conv1d(
            in_channels=d_inner,
            out_channels=d_inner,
            bias=conv_bias,
            kernel_size=d_conv,
            groups=d_inner,
            padding=d_conv - 1,
        )

        self.x_proj = nn.Linear(d_inner, dt_rank + d_state * 2, bias=False)

        self.dt_proj = nn.Linear(dt_rank, d_inner, bias=True)

        A = repeat(torch.arange(1, d_state + 1), 'n -> d n', d=d_inner)
        self.A_log = nn.Parameter(torch.log(A))
        self.D = nn.Parameter(torch.ones(d_inner))
        self.out_proj = nn.Linear(d_inner, d_model, bias=bias)

    def forward(self, x: Tensor) -> Tensor:
        (b, l, d) = x.shape

        x_and_res = self.in_proj(x)  # shape (b, l, 2 * d_in)
        (x, res) = x_and_res.split(split_size=[self.d_inner, self.d_inner], dim=-1)

        x = rearrange(x, 'b l d_in -> b d_in l')
        x = self.conv1d(x)[:, :, :l]
        x = rearrange(x, 'b d_in l -> b l d_in')

        x = F.silu(x)

        y = self.ssm(x)

        y = y * F.silu(res)

        output = self.out_proj(y)

        return output

    def ssm(self, x):
        (d_in, n) = self.A_log.shape

        A = -torch.exp(self.A_log.float())  # shape (d_in, n)
        D = self.D.float()

        x_dbl = self.x_proj(x)  # (b, l, dt_rank + 2*n)

        (delta, B, C) = x_dbl.split(split_size=[self.dt_rank, n, n], dim=-1)  # delta: (b, l, dt_rank). B, C: (b, l, n)
        delta = F.softplus(self.dt_proj(delta))  # (b, l, d_in)

        y = self.selective_scan(x, delta, A, B, C, D)  # This is similar to run_SSM(A, B, C, u) in The Annotated S4 [2]

        return y

    def selective_scan(self, u, delta, A, B, C, D):
        (b, l, d_in) = u.shape
        n = A.shape[1]

        deltaA = torch.exp(einsum(delta, A, 'b l d_in, d_in n -> b l d_in n'))
        deltaB_u = einsum(delta, B, u, 'b l d_in, b l n, b l d_in -> b l d_in n')
        x = torch.zeros((b, d_in, n), device=deltaA.device)
        ys = []
        for i in range(l):
            x = deltaA[:, i] * x + deltaB_u[:, i]
            y = einsum(x, C[:, i, :], 'b d_in n, b n -> b d_in')
            ys.append(y)
        y = torch.stack(ys, dim=1)  # shape (b, l, d_in)

        y = y + u * D

        return y


class RMSNorm(nn.Module):
    def __init__(self,
                 d_model: int,
                 eps: float = 1e-5):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(d_model))

    def forward(self, x):
        output = x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps) * self.weight

        return output


def posemb_sincos_1d(patches: Tensor, temperature: int = 10000) -> Tensor:
    n: int = patches.shape[1]
    d_model: int = patches.shape[2]
    device: torch.device = patches.device
    dtype: torch.dtype = patches.dtype

    if (d_model % 2) != 0:
        raise ValueError(f'feature dimension must be multiple of 2 for sincos emb. got {d_model}')

    n_tensor: Tensor = torch.arange(n, device=device)
    omega: Tensor = torch.arange(d_model // 2, device=device) / (d_model // 2 - 1)
    omega = 1.0 / (temperature ** omega)

    n_tensor = n_tensor.flatten()[:, None] * omega[None, :]
    pe: Tensor = torch.cat((n_tensor.sin(), n_tensor.cos()), dim=1)

    return pe.to(dtype)


class MambaRegressor(Module):
    def __init__(self, *,
                 context_length: int,
                 n_patches: int,
                 d_model: int,
                 d_state: int,
                 d_conv: int,
                 expand: int,
                 n_layers: int,
                 layer_norm: bool = False,
                 conv_bias: bool = False,
                 bias: bool = False,
                 ) -> None:
        super().__init__()

        patch_size: int = context_length // n_patches

        if context_length % patch_size != 0:
            raise ValueError(f"sequence length {context_length} must be divisible by patch size {patch_size}")

        self.pos_embedding = posemb_sincos_1d

        self.to_patch_embedding = Sequential(
            Rearrange('b (n p) -> b n p', p=patch_size),
            LayerNorm(patch_size) if layer_norm else Identity(),
            Linear(patch_size, d_model),
            LayerNorm(d_model) if layer_norm else Identity(),
        )

        self.layers: ModuleList = ModuleList([
            ResidualBlock(
                d_model=d_model,
                d_inner=d_model * expand,
                conv_bias=conv_bias,
                d_conv=d_conv,
                bias=bias,
                d_state=d_state,
                dt_rank=math.ceil(d_model / 16),
            )
            for _ in range(n_layers)
        ])

        self.norm_f = RMSNorm(d_model) if layer_norm else Identity()

        self.linear_head = Sequential(
            LayerNorm(d_model) if layer_norm else Identity(),
            Linear(d_model, 1),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.to_patch_embedding(x)
        x += self.pos_embedding(x)

        for layer in self.layers:
            x = layer(x)

        x = self.norm_f(x)

        x = x.mean(dim=1)

        x = self.linear_head(x).squeeze()

        return x
